/*
 * Copyright 2007 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Niels Provos.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/time.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/queue.h>
#include <sys/tree.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <err.h>
#include <unistd.h>
#include <assert.h>
#include <time.h>

#include <event.h>
#include <evdns.h>
#include <evhttp.h>

#include "spybye.h"
#include "utils.h"
#include "status.h"
#include "log.h"

/* tell our callers that the name could not be resolved */
static void dns_dispatch_error(struct dns_cache *);
static void dns_dispatch_requests(struct dns_cache *dns_entry);
static void inform_domain_notfound(struct evhttp_request *request);
static void inform_proxy_settings(struct evhttp_request *request);
static void inform_no_referer(struct evhttp_request *request);

int debug = 0;

/* globals */

static int log_fd = -1;
static int log_report_fd = -1;
static int allow_private_ip = 0;
static short http_port = 8080;	/* port on which the server runs */

static int
dns_compare(struct dns_cache *a, struct dns_cache *b)
{
	return strcasecmp(a->name, b->name);
}

static SPLAY_HEAD(dns_tree, dns_cache) root;

SPLAY_PROTOTYPE(dns_tree, dns_cache, node, dns_compare);
SPLAY_GENERATE(dns_tree, dns_cache, node, dns_compare);

static void
dns_ttl_expired(int result, short what, void *arg)
{
	struct dns_cache *dns = arg;
	
	fprintf(stderr, "Expiring dns entry for %s\n", dns->name);

	assert(TAILQ_FIRST(&dns->entries) == NULL);
	dns_free(dns);
}

static void
dns_resolv_cb(int result, char type, int count, int ttl,
    void *addresses, void *arg)
{
	struct dns_cache *entry = arg;
	struct timeval tv;

	if (result != DNS_ERR_NONE) {
		/* we were not able to resolve the name */
		dns_dispatch_error(entry);
		return;
	}

	if (!allow_private_ip && check_private_ip(addresses, count)) {
		/* we can't go here - it might be our own machine! */
		dns_dispatch_error(entry);
		return;
	}

	/* copy the addresses */
	entry->addresses = calloc(count, sizeof(struct in_addr));
	if (entry->addresses == NULL)
		err(1, "calloc");
	entry->address_count = count;
	memcpy(entry->addresses, addresses, count * sizeof(struct in_addr));

	dns_dispatch_requests(entry);

	/* expire it after its time-to-live is over */
	evtimer_set(&entry->ev_timeout, dns_ttl_expired, entry);
	timerclear(&tv);
	tv.tv_sec = ttl;
	evtimer_add(&entry->ev_timeout, &tv);
}

static void
http_copy_headers(struct evkeyvalq *dst, struct evkeyvalq *src)
{
	struct evkeyval *kv;
	TAILQ_FOREACH(kv, src, next) {
		if (strcasecmp(kv->key, "Transfer-Encoding") == 0 ||
		    strcasecmp(kv->key, "Connection") == 0 ||
		    strcasecmp(kv->key, "Keep-Alive") == 0 ||
		    strcasecmp(kv->key, "Proxy-Connection") == 0) {
			if (debug)
				fprintf(stderr, "Ignoring %s: %s\n",
				    kv->key, kv->value);
			continue;
		}
		/* we might want to do some filtering here */
		if (debug)
			fprintf(stderr, "Copying %s: %s\n",
			    kv->key, kv->value);
		evhttp_add_header(dst, kv->key, kv->value);
	}
}

static void
map_location_header(struct evhttp_request *req, const char *location)
{
	static char path[1024];
	char *host, *uri;
	u_short port;
			
	if (location[0] == '/') {
		if (http_hostportfile(req->uri, &host, &port, &uri) == -1)
			return;
		snprintf(path, sizeof(path),
		    "http://%s%s", host, location);
	} else {
		strlcpy(path, location, sizeof(path));
	}
	fprintf(stderr, "Mapping %s -> %s\n", path, req->uri);
	site_new(path, req->uri);
}

static void
inform_dangerous(struct evhttp_request *request)
{
	struct evbuffer *databuf = evbuffer_new();
	char *escaped = evhttp_htmlescape(request->uri);
	assert(databuf != NULL);

	evbuffer_add_printf(databuf,
	    "<html><head><title>Dangerous Request</title></head>"
	    "<body>Possibly dangerous load of %s</body></html>",
	    escaped);
	free(escaped);

	/* we cannot allow this request */
	evhttp_send_reply(request,
	    HTTP_NOTFOUND, "Disallowing dangerous request.",
	    databuf);
	evbuffer_free(databuf);
}

static void
inform_domain_notfound(struct evhttp_request *request)
{
	struct evbuffer *databuf = evbuffer_new();
	char *escaped = evhttp_htmlescape(request->uri);
	assert(databuf != NULL);

	evbuffer_add_printf(databuf,
	    "<html><head><title>Domain not found</title></head>"
	    "<body><h1>Domain not found</h1>\n"
	    "Cannot find an IP address for %s</body></html>",
	    escaped);
	free(escaped);

	/* we cannot allow this request */
	evhttp_send_reply(request,
	    HTTP_BADREQUEST, "Disallowing dangerous request.",
	    databuf);
	evbuffer_free(databuf);
}

static void
inform_proxy_settings(struct evhttp_request *request)
{
	struct evbuffer *databuf = evbuffer_new();
	assert(databuf != NULL);

	evbuffer_add_printf(databuf,
	    "<html><head><title>Proxy Configuration</title></head>"
	    "<body><div style=\"border: solid 1px; padding: 2px; "
	    "width: 40%%; "
	    "background-color: #dcdcee; font-family: Verdana, Arial;\">"
	    "<h2>Proxy Configuration</h2>\n"
	    "To use <i>SpyBye</i>, you need to configure your browser to "
	    "use the following proxy <b>http://%s/</b></div><p>"
	    "<iframe src=\"http://www.spybye.org/\" width=100%% height=85%%>"
	    "</body></html>",
	    evhttp_find_header(request->input_headers, "Host"));

	/* we cannot allow this request */
	evhttp_send_reply(request, HTTP_OK, "OK", databuf);
	evbuffer_free(databuf);
}

static void
inform_no_referer(struct evhttp_request *request)
{
	struct evbuffer *databuf = evbuffer_new();
	char *escaped = evhttp_encode_uri(request->uri);
	char *html_escaped = evhttp_htmlescape(request->uri);
	assert(databuf != NULL);

	evbuffer_add_printf(databuf,
	    "<html><head><title>Request Denied</title></head>"
	    "<body><div style=\"border: solid 1px; padding: 2px; "
	    "width: 40%%; "
	    "background-color: #dcdcee; font-family: Verdana, Arial;\">"
	    "<h2>Request Denied</h2>\n"
	    "To use <i>SpyBye</i>, visit "
	    "<a href=\"http://spybye.org/\">http://spybye.org/</a> "
	    "or go to <a href=\"http://spybye.org/?url=%s\">%s</a> "
	    "using SpyBye.</div><p>"
	    "</body></html>",
	    escaped, html_escaped);
	free(escaped);
	free(html_escaped);

	/* we cannot allow this request */
	evhttp_send_reply(request, HTTP_NOTFOUND, "Not Found", databuf);
	evbuffer_free(databuf);
}

static void
http_request_done(struct evhttp_request *req, void *arg)
{
	struct site *site = NULL;
	struct proxy_request *pr = arg;
	const char *location = NULL;

	if (req == NULL || req->response_code == 0 ||
	    (site = site_find(pr->req->uri)) == NULL) {
		evhttp_send_error(pr->req,
		    HTTP_SERVUNAVAIL, "Unknown error");
		proxy_request_free(pr);
		return;
	}

	/* store the data that we received */
	site_cache_data(site, req);

	/* not sanitize some stupid javascript */
	sanitize_content(req);

	/* get the site state for filter for danger */
	if (site->danger == DANGEROUS) {
		inform_dangerous(pr->req);
		return;
	}

	location = evhttp_find_header(req->input_headers, "Location");
	/* keep track of the redirect so that we can tie it together */
	if (location != NULL)
		map_location_header(pr->req, location);
	http_copy_headers(pr->req->output_headers, req->input_headers);

	/* make everything we do no-cacheable */
	evhttp_remove_header(pr->req->output_headers, "Pragma");
	evhttp_add_header(pr->req->output_headers,
	    "Pragma", "no-cache, no-store");

	evhttp_remove_header(pr->req->output_headers, "Cache-Control");
	evhttp_add_header(pr->req->output_headers,
	    "Cache-Control", "no-cache, no-store, must-revalidate, max-age=-1");

	evhttp_send_reply(pr->req,
	    req->response_code, req->response_code_line,
	    req->input_buffer);
	proxy_request_free(pr);
}

static void
dispatch_single_request(struct dns_cache *dns, struct proxy_request *pr)
{
	struct evhttp_request *request;
	char *address = inet_ntoa(dns->addresses[0]);

	assert(pr->evcon == NULL);
	pr->evcon = evhttp_connection_new(address, pr->port);
	fprintf(stderr, "Making connection to %s:%d\n", address, pr->port);
	if (pr->evcon == NULL)
		goto fail;

	/* we got the connection now - queue the request */
	request = evhttp_request_new(http_request_done, pr);
	if (request == NULL)
		goto fail;

	http_copy_headers(request->output_headers, pr->req->input_headers);
	evhttp_add_header(request->output_headers,
	    "X-Forwarded-For", pr->req->remote_host);

	evhttp_add_header(request->output_headers, "Connection", "close");
	evhttp_make_request(pr->evcon, request, EVHTTP_REQ_GET, pr->uri);
	return;

fail:
	evhttp_send_error(pr->req,
	    HTTP_SERVUNAVAIL, "Out of resources");
	proxy_request_free(pr);
	return;
}

static void
dns_dispatch_requests(struct dns_cache *dns)
{
	struct proxy_request *entry;
	while ((entry = TAILQ_FIRST(&dns->entries)) != NULL) {
		TAILQ_REMOVE(&dns->entries, entry, next);
		
		dispatch_single_request(dns, entry);
	}
}

static void
dns_dispatch_error(struct dns_cache *dns_entry)
{
	struct proxy_request *entry;
	while ((entry = TAILQ_FIRST(&dns_entry->entries)) != NULL) {
		TAILQ_REMOVE(&dns_entry->entries, entry, next);

		inform_domain_notfound(entry->req);
		proxy_request_free(entry);
	}

	/* no negative caching */
	dns_free(dns_entry);
}

struct dns_cache *
dns_new(const char *name)
{
	struct dns_cache *entry, tmp;
	struct in_addr address;


	tmp.name = (char *)name;
	if ((entry = SPLAY_FIND(dns_tree, &root, &tmp)) != NULL)
		return (entry);

	entry = calloc(1, sizeof(struct dns_cache));
	if (entry == NULL)
		err(1, "calloc");

	entry->name = strdup(name);
	if (entry->name == NULL)
		err(1, "strdup");

	TAILQ_INIT(&entry->entries);
	SPLAY_INSERT(dns_tree, &root, entry);

	if (inet_aton(entry->name, &address) != 1) {
		evdns_resolve_ipv4(entry->name, 0,
		    dns_resolv_cb, entry);
	} else {
		/* this request is dangerous */
		if (!allow_private_ip && check_private_ip(&address, 1)) {
			dns_free(entry);
			return (NULL);
		}

		/* we already have an address - no dns necessary */
		dns_resolv_cb(DNS_ERR_NONE, DNS_IPv4_A,
		    1, 3600, &address, entry);
	}

	return (entry);
}

void
dns_free(struct dns_cache *entry)
{
	SPLAY_REMOVE(dns_tree, &root, entry);
	free(entry->name);
	free(entry);
}

void
request_add_dns(struct dns_cache *entry, struct proxy_request *pr)
{
	TAILQ_INSERT_TAIL(&entry->entries, pr, next);

	/* still waiting for resolution */
	if (entry->address_count == 0)
		return;

	dns_dispatch_requests(entry);
}

/* keep some simple stats */
static void
spybye_traffic_stats(struct site *site)
{
	extern struct stats statistics;

	statistics.num_requests++;
	switch (site->danger) {
	case HARMLESS:
		statistics.num_harmless++;
		break;
	case DANGEROUS:
		statistics.num_dangerous++;
		break;
	case UNKNOWN:
	default:
		statistics.num_unknown++;
		break;
	}
}

/*
 * Receive all possible requests - analyze them for doing stuff
 */

void
request_handler(struct evhttp_request *request, void *arg)
{
	char *host, *uri;
	const char *referer;
	u_short port;
	struct site *site;
	struct dns_cache *entry;
	struct proxy_request *pr;

	if (request->uri != NULL && request->uri[0] == '/') {
		if (strcmp(request->uri, "/reports") == 0) {
			/* save the report somewhere on disk */
			if (log_report_fd != -1)
				log_report(log_report_fd, request);
		} else {
			inform_proxy_settings(request);
		}
		return;
	}

	if (http_hostportfile(request->uri, &host, &port, &uri) == -1) {
		evhttp_send_error(request,
		    HTTP_BADREQUEST, "Illegal request.");
		return;
	}

	if (strncasecmp(host, "spybye", 6) == 0) {
		fprintf(stderr, "Received request for %s\n", request->uri);
		spybye_handle_request(request, arg);
		return;
	}

	/* now insert the request into our status object */
	referer = evhttp_find_header(request->input_headers, "Referer");
	fprintf(stderr, "Received request for %s (%s) from %s\n",
	    request->uri, referer, request->remote_host);
	if (referer == NULL) {
		inform_no_referer(request);
		return;
	}
	
	site = site_new(request->uri, referer);
	if (site == NULL) {
		/* some evil sites are tricky */
		char *decode = evhttp_decode_uri(referer);
		site = site_new(request->uri, decode);
		free(decode);
	}
	if (site == NULL) {
		fprintf(stderr, "Missing site state for %s\n", request->uri);
		inform_no_referer(request);
		return;
	}

	spybye_traffic_stats(site);

	/*
	 * make requests to dangerous sites just to keep track of the content.
	 * xxx: we might want to wait with logging this until we got the data.
	 */
	if (site->danger == DANGEROUS) {
		if (log_fd != -1)
			log_dangerous_request(log_fd, request);
		if (log_report_fd != -1)
			log_dangerous_report(log_report_fd, request);
	}

	/* make sure that we do not send a referer if this is a root URL */
	if (site->parent == NULL)
		evhttp_remove_header(request->input_headers, "Referer");

	if ((entry = dns_new(host)) == NULL) {
		fprintf(stderr, "Attempt to visit private IP: %s\n",
		    request->uri);
		evhttp_send_error(request,
		    HTTP_BADREQUEST, "Access to private IP disallowed.");
		return;
	}
	pr = proxy_request_new(request, port, uri);
	request_add_dns(entry, pr);
}

struct proxy_request *
proxy_request_new(struct evhttp_request *req, u_short port, char *uri)
{
	struct proxy_request *pr;

	if ((pr = calloc(1, sizeof(struct proxy_request))) == NULL)
		err(1, "calloc");

	pr->uri = strdup(uri);
	if (pr->uri == NULL)
		err(1, "strdup");

	pr->req = req;
	pr->port = port;

	return (pr);
}

void
proxy_request_free(struct proxy_request *pr)
{
	/* if (pr->evcon != NULL)
	   evhttp_connection_free(pr->evcon); */

	free(pr->uri);
	free(pr);
}

void
usage(const char *progname)
{
	fprintf(stderr,
	    "%s: [-P] [-p port] [-g good] [-b bad]\n"
	    "\t -P disable private IP check; allows the proxy to fetch 127/8\n"
	    "\t -g good_patterns a file or url containing the good patterns\n"
	    "\t -b bad_patterns a file or url containing the danger patterns\n"
	    "\t for documentation of all options consult the man page\n",
	    progname);
}

int
main(int argc, char **argv)
{
	extern char *optarg;
	extern int optind;
	struct evhttp *http_server = NULL;
	char *reports = "http://www.spybye.org:8080/reports";
	char *goodness = "http://www.monkey.org/~provos/good_patterns";
	char *badness = "http://www.monkey.org/~provos/bad_patterns";
	char *log_file = "spybye.log";
	char *log_report_file = NULL;
	int ch;

	while ((ch = getopt(argc, argv, "Pp:g:b:l:S:R:")) != -1)
		switch(ch) {
		case 'R':
			log_report_file = optarg;
			break;
		case 'S':
			reports = optarg;
			break;
		case 'P':
			allow_private_ip = 1;
			break;
		case 'p':
			http_port = atoi(optarg);
			if (!http_port) {
				usage(argv[0]);
				exit(1);
			}
			break;
		case 'l':
			if (!strlen(optarg))
				log_file = NULL;
			else
				log_file = optarg;
			break;
		case 'g':
			goodness = optarg;
			break;
		case 'b':
			badness = optarg;
			break;
		default:
			usage(argv[0]);
			exit(1);
		}

	argc -= optind;
	argv += optind;

	event_init();
	evtag_init();
	evdns_init();

	fprintf(stderr, "SpyBye %s starting up ...\n", VERSION);
	
	if (strlen(reports)) {
		extern struct spybye_share spybye_share;
		log_establish_sharing(&spybye_share, reports);
	}

	status_init(goodness, badness);

	/* open the log file if needed */
	if (log_file != NULL) {
		log_dangerous_read(log_file);
		log_fd = log_init(log_file);
	}

	if (log_report_file != NULL) {
		log_report_fd = log_init(log_report_file);
	}

	http_server = evhttp_start("0.0.0.0", http_port);
	if (http_server == NULL) {
		fprintf(stderr, "Cannot run web server on port %d\n",
		    http_port);
		exit(1);
	}
	evhttp_set_gencb(http_server, request_handler, NULL);

	fprintf(stderr, "Starting web server on port %d\n"
	    "Configure your browser to use this server as proxy.\n",
	    http_port);

	SPLAY_INIT(&root);

	event_dispatch();

	if (log_fd != -1)
		log_close(log_fd);

	exit (0);
}
